import threading
from nanolab import pycmd
from .base_mixin import CommandMixinBase
from collections import defaultdict
import heapq
import time


class ThreadedCommandMixin(CommandMixinBase):

    def _execute_command_sequence(self, commands):
        for command_config in commands:
            self.command_instance.execute_another_command(command_config)

    def validate(self):

        for python_command in self.command_instance.command_config["commands"]:
            if "method" not in python_command:
                raise ValueError(f"Python command: 'method' is required.")

            method_name = python_command["method"]

            if "class" in python_command:
                cls = getattr(pycmd, python_command["class"])
                if not cls:
                    raise ValueError(
                        f"Python command: Class '{python_command['class']}' not found."
                    )
                if not hasattr(cls, method_name):
                    raise ValueError(
                        f"Python command: Method '{method_name}' not found in class '{python_command['class']}'."
                    )
            else:
                if method_name not in globals():
                    raise ValueError(
                        f"Python command: Method '{method_name}' not found in the current module."
                    )

    def execute(self):
        schedule = []
        command_groups = defaultdict(list)
        config = self.command_instance.command_config["commands"]

        min_delay = min(cmd.get("delay", 0) for cmd in config)
        # Normalize all delays by adding the absolute value of the minimum delay
        config = [{
            **cmd, "delay": cmd.get("delay", 0) + abs(min_delay)
        } for cmd in config]

        def add_to_schedule(delay, target, args):
            # Get the current time and add the delay (could be negative)
            run_at = time.time() + delay
            # Add the task to the schedule
            heapq.heappush(schedule, (run_at, target, args))

        for command_config in config:
            group = command_config.get("group")
            if group:
                command_groups[group].append(command_config)
            else:
                delay = command_config.get("delay", 0)
                add_to_schedule(delay,
                                self.command_instance.execute_another_command,
                                (command_config, ))

        for group, command_configs in command_groups.items():
            delay = command_configs[0].get("delay", 0)
            add_to_schedule(delay, self._execute_command_sequence,
                            (command_configs, ))

        threads = []
        while schedule:
            run_at, target, args = heapq.heappop(schedule)
            delay = max(0, run_at - time.time())
            time.sleep(delay)  # wait until it's time to start this task
            thread = threading.Thread(target=target, args=args)
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()



# nanolab/command/argparse_commands.py

from os import environ
from .github import GitHubAPI
import argparse
from nanolab.src.config_handler import ConfigPathHandler, ConfigResourceHandler
from nanomock.modules.nl_parse_config import ConfigReadWrite
from nanolab.src.utils import extract_packaged_data_to_disk
from nanolab.src.config_loader import ConfigLoader, ConfigValidator, ConfigCommandExecutor
from nanolab.src.snippet_manager import SnippetManager


def parse_args():
    parser = argparse.ArgumentParser(
        description="Load configuration and snippets files")

    subparsers = parser.add_subparsers(dest="command")

    # Run command
    run_parser = subparsers.add_parser("run", help="Run a testcase")
    run_parser.add_argument('-t', '--testcase', type=str, help='Testcase name')
    run_parser.add_argument('-i', '--image', nargs='+', help='Docker image(s)')
    run_parser.add_argument('--gh-user',
                            type=str,
                            default='gr0vity-dev',
                            help='GitHub user (default: gr0vity-dev)')
    run_parser.add_argument(
        '--gh-repo',
        type=str,
        default='nanolab-configs',
        help='GitHub repository (default: nanolab-configs)')
    run_parser.add_argument(
        '--gh-path',
        type=str,
        default='default',
        help='Path to testcases within the repository (default: default)')

    # List command
    list_parser = subparsers.add_parser("list", help="List testcases")
    list_parser.add_argument('--gh-user',
                             type=str,
                             default='gr0vity-dev',
                             help='GitHub user (default: gr0vity-dev)')
    list_parser.add_argument(
        '--gh-repo',
        type=str,
        default='nanolab-configs',
        help='GitHub repository (default: nanolab-configs)')
    list_parser.add_argument(
        '--gh-path',
        type=str,
        default='default',
        help='Path to testcases within the repository (default: default)')

    return parser.parse_args()


class ArgParseHandler:

    def __init__(self, args=None):
        self.args = args or parse_args()
        self.conf_rw = ConfigReadWrite()
        self.github_api = GitHubAPI(self.args.gh_user, self.args.gh_repo,
                                    self.args.gh_path)

    def run(self):
        path_handler = ConfigPathHandler(self.args.testcase)
        resource_handler = ConfigResourceHandler(self.github_api, path_handler)

        #
        extract_packaged_data_to_disk(path_handler.get_snippets_path())

        # copy config to disk
        config_path = resource_handler.copy_config_file(
            path_handler.config_copy_file_path)

        # resolve and downlaod resources defiend in the config
        resolved_config = resource_handler.download_and_replace_resources(
            config_path, path_handler.downloads_path)
        # replace docker_tags with commandline
        if self.args.image:
            resolved_config["docker_tags"] = self.args.image

        resolved_path = path_handler.get_resolved_config_path()
        snippet_path = path_handler.get_snippets_path()

        snippet_manager = SnippetManager(snippet_path)
        config_loader = ConfigLoader(snippet_manager)
        final_config = config_loader.load_config(resolved_config)
        self.conf_rw.write_json(resolved_path, final_config)
        ConfigValidator.validate_config(final_config)
        ConfigCommandExecutor.execute_commands(final_config)

    def list_testcases(self):
        files = self.github_api.get_files()

        if files:
            for file in files:
                if file["name"].endswith(".json"):
                    print(file["name"][:-5])

    def get_args(self):
        return self.args
